import * as plugins from './plugins.js';
import * as interfaces from './env.js';
import { logger } from './logging.js';
import { ServiceWorker } from './classes.serviceworker.js';

export class CacheManager {
  public losslessServiceWorkerRef: ServiceWorker;

  public usedCacheNames = {
    runtimeCacheName: 'runtime'
  };

  /**
   * Hard cached domains are always attempted to be cached, regardless of browser.
   * For example, your internal origin, fonts, CDNs, etc.
   */
  public hardCachedDomains: string[];

  /**
   * Soft cached domains will be cached normally except on Safari.
   * This is useful for domains where Safari’s handling of cached CORS responses is problematic.
   */
  public softCachedDomains: string[];

  constructor(losslessServiceWorkerRefArg: ServiceWorker) {
    this.losslessServiceWorkerRef = losslessServiceWorkerRefArg;

    // Default hard cached domains.
    this.hardCachedDomains = [
      this.losslessServiceWorkerRef.serviceWindowRef.location.origin,
      'https://unpkg.com',
      'https://fonts.googleapis.com',
      'https://fonts.gstatic.com'
    ];

    // Default soft cached domains.
    this.softCachedDomains = [
      'https://assetbroker.'
    ];

    this._setupCache();
  }

  /**
   * Returns true if the given URL is in one of the hard cached domains.
   */
  private isHardCached(url: string): boolean {
    return this.hardCachedDomains.some(domain => url.includes(domain));
  }

  /**
   * Returns true if the given URL is in one of the soft cached domains.
   */
  private isSoftCached(url: string): boolean {
    return this.softCachedDomains.some(domain => url.includes(domain));
  }

  /**
   * Returns true if the given URL is cacheable (i.e. belongs to either hard or soft domains).
   */
  private isCacheable(url: string): boolean {
    return this.isHardCached(url) || this.isSoftCached(url);
  }

  /**
   * Sets up the fetch event listener and caching logic.
   */
  private _setupCache = () => {
    const createMatchRequest = (requestArg: Request): Request => {
      let matchRequest: Request;
      // For internal requests, use the original request.
      if (requestArg.url.startsWith(this.losslessServiceWorkerRef.serviceWindowRef.location.origin)) {
        matchRequest = requestArg;
      } else {
        // For external requests, create a new Request with appropriate CORS settings.
        // For soft cached domains, we use 'no-cors' to keep the response opaque.
        const isSoft = this.isSoftCached(requestArg.url);
        const mode: RequestMode = isSoft ? 'no-cors' : 'cors';
        matchRequest = new Request(requestArg.url, {
          method: requestArg.method,
          headers: requestArg.headers,
          mode,
          credentials: 'same-origin',
          redirect: 'follow'
        });
      }
      return matchRequest;
    };

    /**
     * Creates a 500 error response.
     */
    const create500Response = async (requestArg: Request, responseArg: Response): Promise<Response> => {
      return new Response(
        `
        <html>
          <head>
            <style>
              .note {
                padding: 10px;
                color: #fff;
                background: #000;
                border-bottom: 1px solid #e4002b;
                margin-bottom: 20px;
              }
            </style>
          </head>
          <body>
            <div class="note">
              <strong>Service worker running, but status 500</strong><br>
            </div>
            Service worker is unable to fetch this request.<br>
            Request URL: ${requestArg.url}<br>
            Response Type: ${responseArg.type}<br>
            Response Body: ${await responseArg.clone().text()}<br>
          </body>
        </html>
      `,
        {
          headers: { "Content-Type": "text/html" },
          status: 500
        }
      );
    };

    this.losslessServiceWorkerRef.serviceWindowRef.addEventListener('fetch', async (fetchEventArg: any) => {
      const originalRequest: Request = fetchEventArg.request;
      const parsedUrl = new URL(originalRequest.url);

      // Exclude specific hosts or paths from being handled by the service worker.
      if (
        parsedUrl.hostname.includes('paddle.com') ||
        parsedUrl.hostname.includes('paypal.com') ||
        parsedUrl.hostname.includes('reception.lossless.one') ||
        parsedUrl.pathname.startsWith('/socket.io') ||
        originalRequest.url.startsWith('https://umami.')
      ) {
        logger.log('note', `Service worker not active for ${parsedUrl.toString()}`);
        return;
      }

      // Create a deferred response.
      const done = plugins.smartpromise.defer<Response>();
      fetchEventArg.respondWith(done.promise);

      // Determine if the request should be cached.
      if (originalRequest.method === 'GET' && this.isCacheable(originalRequest.url)) {
        // Check if running on Safari.
        const userAgent = (self.navigator && self.navigator.userAgent) || "";
        const isSafari = /Safari/.test(userAgent) && !/Chrome/.test(userAgent);

        // For soft cached domains on Safari, bypass caching.
        if (this.isSoftCached(originalRequest.url) && isSafari) {
          logger.log('info', `Safari detected – bypass caching for soft cached domain: ${originalRequest.url}`);
          const networkResponse = await fetch(originalRequest).catch(async err => {
            return await create500Response(originalRequest, new Response(err.message));
          });
          done.resolve(networkResponse);
          return;
        }

        // For other cases, try serving from cache.
        const matchRequest = createMatchRequest(originalRequest);
        const cachedResponse = await caches.match(matchRequest);
        if (cachedResponse) {
          logger.log('ok', `CACHED: found cached response for ${matchRequest.url}`);
          done.resolve(cachedResponse);
          return;
        }

        logger.log('info', `NOTYETCACHED: fetching and caching ${matchRequest.url}`);
        const newResponse: Response = await fetch(matchRequest).catch(async err => {
          return await create500Response(matchRequest, new Response(err.message));
        });

        // Do not cache responses that are not successful or are opaque (when not expected).
        if (newResponse.status > 299 || newResponse.type === 'opaque') {
          logger.log(
            'error',
            `NOTCACHED: cannot cache response for ${matchRequest.url} (status: ${newResponse.status}, type: ${newResponse.type})`
          );
          done.resolve(await create500Response(matchRequest, newResponse));
          return;
        } else {
          // Open the runtime cache.
          const cache = await caches.open(this.usedCacheNames.runtimeCacheName);
          const responseToCache = newResponse.clone();

          // Build new headers preserving all except caching-related headers.
          const headers = new Headers();
          responseToCache.headers.forEach((value, key) => {
            if (!['Cache-Control', 'cache-control', 'Expires', 'expires', 'Pragma', 'pragma'].includes(key)) {
              headers.set(key, value);
            }
          });

          // Ensure necessary CORS headers are present.
          if (!headers.has('Access-Control-Allow-Origin')) {
            headers.set('Access-Control-Allow-Origin', '*');
          }
          if (!headers.has('Access-Control-Allow-Methods')) {
            headers.set('Access-Control-Allow-Methods', 'GET, POST, PUT, DELETE, OPTIONS');
          }
          if (!headers.has('Access-Control-Allow-Headers')) {
            headers.set('Access-Control-Allow-Headers', 'Content-Type');
          }
          // Set caching headers to prevent browser caching.
          headers.set('Cache-Control', 'no-store, no-cache, must-revalidate, proxy-revalidate');
          headers.set('Pragma', 'no-cache');
          headers.set('Expires', '0');
          headers.set('Surrogate-Control', 'no-store');

          // Read the response body as a blob (helps prevent issues with locked streams on Safari).
          let newCachedResponse: Response;
          try {
            const bodyBlob = await responseToCache.blob();
            newCachedResponse = new Response(bodyBlob, {
              status: responseToCache.status,
              statusText: responseToCache.statusText,
              headers
            });
          } catch (err) {
            newCachedResponse = newResponse;
          }

          await cache.put(matchRequest, newCachedResponse);
          logger.log('ok', `NOWCACHED: response for ${matchRequest.url} cached successfully`);
          done.resolve(newResponse);
        }
      } else {
        // For non-cacheable requests, fetch directly from the network.
        logger.log('ok', `NOTCACHED: fetching ${originalRequest.url} from origin (non-cacheable)`);
        const networkResponse = await fetch(originalRequest).catch(async err => {
          return await create500Response(originalRequest, new Response(err.message));
        });
        done.resolve(networkResponse);
      }
    });
  };

  /**
   * Cleans all caches.
   * Should only be run when a new service worker is activated.
   * @param reasonArg A reason for the cache cleanup.
   */
  public cleanCaches = async (reasonArg = 'no reason given'): Promise<void> => {
    logger.log('info', `MAJOR CACHEEVENT: cleaning caches now! Reason: ${reasonArg}`);
    const cacheNames = await caches.keys();

    const deletePromises = cacheNames.map(cacheToDelete => {
      const deletePromise = caches.delete(cacheToDelete);
      deletePromise.then(() => {
        logger.log('ok', `Deleted cache ${cacheToDelete}`);
      });
      return deletePromise;
    });
    await Promise.all(deletePromises);
  };

  /**
   * Revalidates the runtime cache.
   */
  public async revalidateCache(): Promise<void> {
    const runtimeCache = await caches.open(this.usedCacheNames.runtimeCacheName);
    const cacheKeys = await runtimeCache.keys();
    for (const requestArg of cacheKeys) {
      const clonedRequest = requestArg.clone();
      const response = await plugins.smartpromise.timeoutWrap(fetch(clonedRequest), 5000);
      if (response && response.status >= 200 && response.status < 300) {
        await runtimeCache.delete(requestArg);
        await runtimeCache.put(requestArg, response);
      }
    }
  }
}